from typing import Union

from pdlpy.decorators import iterable
from pdlpy.distribution import Distribution
from pdlpy.math import e, erf, pi, sqrt


class Normal(Distribution):
    """
    Continuous probability distribution of a random variable X
    which is assumed to be additively produced by many small events
    """

    def __init__(self, e: float, v: float):
        """
        Parameters
        e: the expectation of the distribution
        v: the variance of the distribution
        """
        self._e = e
        self._v = v
        self._mean = e
        self._median = e
        self._mode = e
        self._var = v

    def __str__(self):
        return (
            "Normal("
            f"e={self._e:.2f}, "
            f"v={self._v:.2f}, "
            f"mean={self._mean:.2f}, "
            f"median={self._median:.2f}, "
            f"mode={self._mode:.2f}, "
            f"var={self._var:.2f}"
            ")"
        )

    @iterable
    def pdf(self, x: Union[list, tuple, float]) -> Union[list, tuple, float]:
        """
        Probability Density Function

        Parameters
        x: the value of random variable X

        Returns
        The relative likelihood that X will lie in the sample space of x
        """
        return (1 / sqrt(2 * pi * self._v)) * e ** (-((x - self._e) ** 2 / 2 * self._v))

    @iterable
    def cdf(self, x: Union[list, tuple, float]) -> Union[list, tuple, float]:
        """
        Cumulative Distribution Function

        Parameters
        x: the value of the random variable X

        Returns
        The probability that X will take a value less than or equal to x
        """
        return (1 + erf((x - self._e) / (sqrt(self._v * 2)))) / 2



from typing import Union

from pdlpy.decorators import iterable
from pdlpy.distribution import Distribution
from pdlpy.math import ceil, log2


class Geometric(Distribution):
    """
    Discrete probability distributions of a random variable X
    which defines the number of Bernoulli trials needed to get
    a single success
    """

    def __init__(self, p: float):
        """
        Parameters
        p: the probability of success
        """

        if p < 0 or p > 1:
            raise ValueError("p must be in the range [0, 1]")

        self._p = p
        self._mean = 1 / p
        self._median = ceil(-1 / log2(1 - p)) if p != 1 else 1
        self._mode = 1
        self._var = (1 - p) / (p**2)

    def __str__(self):
        return (
            "Geometric("
            f"p={self._p:.2f}, "
            f"mean={self._mean:.2f}, "
            f"median={self._median:.2f}, "
            f"mode={self._mode:.2f}, "
            f"var={self._var:.2f}"
            ")"
        )

    @iterable
    def pmf(self, x: Union[list, tuple, int]) -> Union[list, tuple, float]:
        """
        Probability Mass Function

        Parameters
        x: the value of the random variable X

        Returns
        The probability that X will take a value exactly equal to x
        """

        if x < 1:
            raise ValueError("x must be greater than or equal to 1")

        return (1 - self._p) ** (x - 1) * self._p

    @iterable
    def cdf(self, x: Union[list, tuple, int]) -> Union[list, tuple, float]:
        """
        Cumulative Distribution Function

        Parameters
        x: the value of the random variable X

        Returns
        The probability that X will take a value less than or equal to x
        """

        if x < 1:
            raise ValueError("x must be greater than or equal to 1")

        if x == 1:
            return self.pmf(1)
        else:
            return self.pmf(x) + self.cdf(x - 1)



from typing import Union

from pdlpy.decorators import iterable
from pdlpy.distribution import Distribution


class Uniform(Distribution):
    """
    Continuous probability distribution of a random variable X
    in an interval where any value of X has an equal probability
    """

    def __init__(self, a: float, b: float):
        """
        Parameters
        a: the minimum value of X
        b: the maximum value of X
        """

        if a >= b:
            raise ValueError("a must be less than b")

        self._a = a
        self._b = b
        self._mean = (a + b) / 2
        self._median = (a + b) / 2
        self._mode = (a + b) / 2
        self._var = (b - a) ** 2 / 12

    def __str__(self):
        return (
            "Uniform("
            f"a={self._a:.2f}, "
            f"b={self._b:.2f}, "
            f"mean={self._mean:.2f}, "
            f"median={self._median:.2f}, "
            f"mode={self._mode:.2f}, "
            f"var={self._var:.2f}"
            ")"
        )

    @iterable
    def pdf(self, x: Union[list, tuple, float]) -> Union[list, tuple, float]:
        """
        Probability Density Function

        Parameters
        x: the value of random variable X

        Returns
        The relative likelihood that X will lie in the sample space of x
        """

        if x < self._a:
            return 0.0

        if x > self._b:
            return 0.0

        return 1 / (self._b - self._a)

    @iterable
    def cdf(self, x: Union[list, tuple, float]) -> Union[list, tuple, float]:
        """
        Cumulative Distribution Function

        Parameters
        x: the value of the random variable X

        Returns
        The probability that X will take a value less than or equal to x
        """

        if x <= self._a:
            return 0.0

        if x >= self._b:
            return 1.0

        return (x - self._a) / (self._b - self._a)



from typing import Union

from pdlpy.decorators import iterable
from pdlpy.distribution import Distribution
from pdlpy.math import ncr


class Hypergeometric(Distribution):
    """
    Discrete probability distribution of a random variable X
    that describes the probability of successes in draws
    from a finite test set
    """

    def __init__(self, n: int, N: int, M: int):
        """
        Parameters
        n: the number of draws
        N: the size of the test set
        M: the number of successes
        """

        if N < 0:
            raise ValueError("N must be greater than or equal to 0")

        if n < 0 or n > N:
            raise ValueError("n must be in the range [0, N]")

        if M < 0 or M > N:
            raise ValueError("M must be in the range [0, N]")

        self._n = n
        self._N = N
        self._M = M
        self._mean = n * M / N
        self._median = 0
        self._mode = 0
        self._var = n * M / N * (1 - M / N) * (N - n) / (N - 1)

    def __str__(self):
        return (
            "Hypergeometric("
            f"n={self._n}, "
            f"N={self._N}, "
            f"M={self._M}, "
            f"mean={self._mean:.2f}, "
            f"median={self._median:.2f}, "
            f"mode={self._mode:.2f}, "
            f"var={self._var:.2f}"
            ")"
        )

    @iterable
    def pmf(self, x: Union[list, tuple, int]) -> Union[list, tuple, float]:
        """
        Probability Mass Function

        Parameters
        x: the value of the random variable X

        Returns
        The probability that X will take a value exactly equal to x
        """

        if x < 0 or x > self._n:
            raise ValueError("x must be in the range [0, n]")

        return (
            ncr(self._M, x)
            * ncr(self._N - self._M, self._n - x)
            / ncr(self._N, self._n)
        )

    @iterable
    def cdf(self, x: Union[list, tuple, int]) -> Union[list, tuple, float]:
        """
        Cumulative Distribution Function

        Parameters
        x: the value of the random variable X

        Returns
        The probability that X will take a value less than or equal to x
        """

        if x < 0 or x > self._n:
            raise ValueError("x must be in the range [0, n]")

        if x == 0:
            return self.pmf(0)
        else:
            return self.pmf(x) + self.cdf(x - 1)



from typing import Union

from pdlpy.decorators import iterable
from pdlpy.distribution import Distribution
from pdlpy.math import e, log


class Exponential(Distribution):
    """
    Continuous probability distribution of a random variable X
    which defines the time between events in a Poisson process
    """

    def __init__(self, r: float):
        """
        Parameters
        r: the average number of events
        """

        if r <= 0:
            raise ValueError("r must be greater than 0")

        self._r = r
        self._mean = r**-1
        self._median = log(2) / r
        self._mode = 0
        self._var = r**-2

    def __str__(self):
        return (
            "Exponential("
            f"r={self._r:.2f}, "
            f"mean={self._mean:.2f}, "
            f"median={self._median:.2f}, "
            f"mode={self._mode:.2f}, "
            f"var={self._var:.2f}"
            ")"
        )

    @iterable
    def pdf(self, x: Union[list, tuple, float]) -> Union[list, tuple, float]:
        """
        Probability Density Function

        Parameters
        x: the value of random variable X

        Returns
        The relative likelihood that X will lie in the sample space of x
        """

        if x < 0:
            raise ValueError("x must be greater than or equal to 0")

        return self._r * e ** (-self._r * x)

    @iterable
    def cdf(self, x: Union[list, tuple, float]) -> Union[list, tuple, float]:
        """
        Cumulative Distribution Function

        Parameters
        x: the value of the random variable X

        Returns
        The probability that X will take a value less than or equal to x
        """

        if x < 0:
            raise ValueError("x must be greater than or equal to 0")

        return 1 - e ** (-self._r * x)



from functools import wraps


def iterable(func):
    @wraps(func)
    def wrapper(self, arg):
        if isinstance(arg, int):
            return func(self, arg)

        if isinstance(arg, float):
            return func(self, arg)

        if isinstance(arg, list):
            return [func(self, x) for x in arg]

        if isinstance(arg, tuple):
            return (func(self, x) for x in arg)

        raise TypeError("Unsupported data type")

    return wrapper



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="pdlpy",
    version="0.4.0",
    author="André Bienemann",
    author_email="andre.bienemann@gmail.com",
    description="Probability Distribution Library for Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/andrebienemann/pdlpy",
    extras_require={
        "dev": ["black", "coverage", "isort", "twine", "wheel"],
        "docs": ["mkdocs", "mkdocs-material"],
    },
    packages=setuptools.find_packages(),
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Scientific/Engineering :: Mathematics",
    ],
)



from typing import Union

from pdlpy.decorators import iterable
from pdlpy.distribution import Distribution
from pdlpy.math import e, factorial, floor


class Poisson(Distribution):
    """
    Discrete probability distribution of a random variable X
    which defines the probability of a given number of events
    occurring in a fixed interval of time or space
    """

    def __init__(self, r: float):
        """
        Parameters
        r: the average number of events
        """

        if r <= 0:
            raise ValueError("r must be greater than 0")

        self._r = r
        self._mean = r
        self._median = floor(r + 1 / 3 - 0.02 / r)
        self._mode = floor(r)
        self._var = r

    def __str__(self):
        return (
            "Poisson("
            f"r={self._r:.2f}, "
            f"mean={self._mean:.2f}, "
            f"median={self._median:.2f}, "
            f"mode={self._mode:.2f}, "
            f"var={self._var:.2f}"
            ")"
        )

    @iterable
    def pmf(self, x: Union[list, tuple, int]) -> Union[list, tuple, float]:
        """
        Probability Mass Function

        Parameters
        x: the value of the random variable X

        Returns
        The probability that X will take a value exactly equal to x
        """

        if x < 0:
            raise ValueError("x must be greater than or equal to 0")

        return (self._r**x) * (e ** (-self._r)) / factorial(x)

    @iterable
    def cdf(self, x: Union[list, tuple, int]) -> Union[list, tuple, float]:
        """
        Cumulative Distribution Function

        Parameters
        x: the value of the random variable X

        Returns
        The probability that X will take a value less than or equal to x
        """

        if x < 0:
            raise ValueError("x must be greater than or equal to 0")

        if x == 0:
            return self.pmf(0)
        else:
            return self.pmf(x) + self.cdf(x - 1)
